import csv
import datetime

import heapq
from collections import namedtuple, defaultdict, deque


Publication = namedtuple('Publication', ['lastname', 'firstname', 'year', 'title', 'publisher']) 


class Repository(object):

    def __init__(self):
        
        self.max_publications = 6
        self.reject_threshold = 5 # years
        self.current_year = datetime.datetime.now().year
        self.items = defaultdict(list)  # a dict of author publications
        self.authors = []         # a list of author names (sorted)
    
    
    def load_csv(self, filename):
    
        with open(filename, 'r') as fh:
            reader = csv.DictReader(fh)
            for row in reader:
                lastname = row['AUTHOR SURNAME'].strip()
                firstname = row['AUTHOR NAME'].strip()
                year = int(row['YEAR OF PUBLICATION'])
                title = row['TITLE'].strip()
                publisher = row['PUBLISHER'].strip()
                                          
                publication = Publication(lastname, firstname, year, title, publisher)
                self.add(publication)
      
        
    def add(self, publication):
        
        # publication is too old
        if self.current_year - publication.year > self.reject_threshold:
            return False
        
        fullname = (publication.lastname, publication.firstname)
        # author is not present in the repository yet
        if fullname not in self.authors:
            heapq.heappush(self.authors, fullname)
        self.items[fullname].append(publication)
        # ensure the author doesn't have more than self.max_publications
        self.items[fullname] = self.get_publications_list(fullname)
        
        return publication in self.items[fullname]
      
        
    def delete(self, publication):
        fullname = (publication.lastname, publication.firstname)
        # no such author
        if fullname not in self.authors:
            return False
        # publication not found
        if publication not in self.items[publication]:
            return False
        
        self.items[fullname].remove(publication)
        
        # remove an author if their last publication have just been removed
        if not self.items[fullname]: 
            self.authors.remove(fullname)
            # restore sorted order
            heapq.heapify(self.authors)
            
        return True
        
        
    def get_publications_list(self, fullname):
        return heapq.nlargest(self.max_publications, self.items[fullname], key=lambda x: x.year)
        
        
    def get_authors_list(self):
        return heapq.nsmallest(len(self.authors), self.authors)
        
        
